@@ -21,11 +21,10 @@ from scipy.stats import linregress
 import tables
 warnings.simplefilter('ignore', category=tables.NaturalNameWarning)
 
+import autopilot
 from autopilot import prefs
 from autopilot.core.loggers import init_logger
 
-
-
 if __name__ == '__main__':
     # Parse arguments - this should have been called with a .json prefs file passed
     # We'll try to look in the default location first
@@ -52,9 +51,8 @@ if __name__ == '__main__':
         else:
             from autopilot.stim.sound import jackclient
 
-from autopilot.core.networking import Pilot_Station, Net_Node, Message
+from autopilot.networking import Message, Net_Node, Pilot_Station
 from autopilot import external
-from autopilot import tasks
 from autopilot.hardware import gpio
 
 
@@ -184,7 +182,8 @@ class Pilot:
             'PARAM': self.l_param, # A parameter is being changed
             'CALIBRATE_PORT': self.l_cal_port, # Calibrate a water port
             'CALIBRATE_RESULT': self.l_cal_result, # Compute curve and store result
-            'BANDWIDTH': self.l_bandwidth # test our bandwidth
+            'BANDWIDTH': self.l_bandwidth, # test our bandwidth
+            'STREAM_VIDEO': self.l_stream_video
         }
 
         # spawn_network gives us the independent message-handling process
@@ -206,6 +205,10 @@ class Pilot:
             for pin in prefs.get('PULLDOWNS'):
                 self.pulls.append(gpio.Digital_Out(int(pin), pull='D', polarity=1))
 
+        # store some hardware we use outside of a task
+        self.hardware = {}
+
+
         self.logger.debug('pullups and pulldowns set')
         # check if the calibration file needs to be updated
 
@@ -256,7 +259,7 @@ class Pilot:
 
         # TODO: Report any calibrations that we have
 
-        hello = {'pilot':self.name, 'ip':self.ip, 'state':self.state}
+        hello = {'pilot':self.name, 'ip':self.ip, 'state':self.state, 'prefs': prefs.get()}
 
         self.node.send(self.parentid, 'HANDSHAKE', value=hello)
 
@@ -266,14 +269,14 @@ class Pilot:
         our Station object will cache this and will handle any
         future requests.
         """
-        self.node.send(self.parentid, 'STATE', self.state, flags={'NOLOG':True})
+        self.node.send(self.name, 'STATE', self.state, flags={'NOLOG':True})
 
     def l_start(self, value):
         """
         Start running a task.
 
         Get the task object by using `value['task_type']` to select from
-        :data:`.tasks.TASK_LIST` , then feed the rest of `value` as kwargs
+        :func:`autopilot.get_task()` , then feed the rest of `value` as kwargs
         into the task object.
 
         Calls :meth:`.autopilot.run_task` in a new thread
@@ -294,9 +297,9 @@ class Pilot:
         try:
             # Get the task object by its type
             if 'child' in value.keys():
-                task_class = tasks.CHILDREN_LIST[value['task_type']]
+                task_class = autopilot.get('children', value['task_type'])
             else:
-                task_class = tasks.TASK_LIST[value['task_type']]
+                task_class = autopilot.get_task(value['task_type'])
             # Instantiate the task
             self.stage_block.clear()
 
@@ -512,6 +515,60 @@ class Pilot:
         #self.networking.set_logging(True)
         #self.node.do_logging.set()
 
+    def l_stream_video(self, value):
+        """
+        Start or stop video streaming
+
+        Args:
+            value (dict): a dictionary of the form::
+
+                {
+                    'starting': bool, # whether we're starting (True) or stopping
+                    'camera': str, # the camera to start/stop, of form 'group.camera_id'
+                    'stream_to': node id that the camera should send to
+                }
+        """
+
+        starting = value.get('starting', False)
+        camera = value.get('camera', None)
+        stream_to = value.get('stream_to', None)
+
+        if camera is None or stream_to is None:
+            self.logger.exception('Need a camera and a place to stream it to!')
+            return
+
+        try:
+            cam_group, cam_id = camera.split('.')
+        except ValueError:
+            self.logger.exception(f'Expected camera id in form group.camera_id, got {camera}')
+            return
+
+        if starting:
+            if cam_group in prefs.get('HARDWARE') and cam_id in prefs.get('HARDWARE')[cam_group]:
+                cam_prefs = prefs.get('HARDWARE')[cam_group][cam_id]
+                cam_obj = get_hardware_class(cam_prefs['type'])(**cam_prefs)
+                if cam_group not in self.hardware.keys():
+                    self.hardware[cam_group] = {}
+                self.hardware[cam_group][cam_id] = cam_obj
+
+                cam_obj.stream(to=stream_to, min_size=1)
+                cam_obj.capture()
+                self.logger.info(f'Starting to stream video from {camera} to {stream_to}')
+
+            else:
+                self.logger.exception(f'No camera in group {cam_group} and id {cam_id} is configured in prefs')
+
+        else:
+            # stopping!
+            if cam_group in self.hardware.keys() and cam_id in self.hardware[cam_group]:
+                cam_obj = self.hardware[cam_group][cam_id]
+                cam_obj.stop()
+                cam_obj.release()
+                del self.hardware[cam_group][cam_id]
+                self.logger.info(f'Stopped streaming camera {camera}')
+            else:
+                self.logger.exception(f'No camera was capturing with group {cam_group} and id {cam_id}, have hardware {self.hardware}')
+
 
 
     def calibration_curve(self, path=None, calibration=None):
@@ -691,44 +748,55 @@ class Pilot:
 
         # TODO: Init sending continuous data here
         self.logger.debug('Starting task loop')
-        while True:
-            # Calculate next stage data and prep triggers
-            data = next(self.task.stages)() # Double parens because next just gives us the function, we still have to call it
-            self.logger.debug('called stage method')
+        try:
+            while True:
+                # Calculate next stage data and prep triggers
+                data = next(self.task.stages)() # Double parens because next just gives us the function, we still have to call it
+                self.logger.debug('called stage method')
 
-            if data:
-                data['pilot'] = self.name
-                data['subject'] = self.subject
+                if data:
+                    data['pilot'] = self.name
+                    data['subject'] = self.subject
 
-                # Send data back to terminal (subject is identified by the networking object)
-                self.node.send('T', 'DATA', data)
+                    # Send data back to terminal (subject is identified by the networking object)
+                    self.node.send('T', 'DATA', data)
 
-                # Store a local copy
-                # the task class has a class variable DATA that lets us know which data the row is expecting
-                if trial_data:
-                    for k, v in data.items():
-                        if k in self.task.TrialData.columns.keys():
-                            row[k] = v
+                    # Store a local copy
+                    # the task class has a class variable DATA that lets us know which data the row is expecting
+                    if trial_data:
+                        for k, v in data.items():
+                            if k in self.task.TrialData.columns.keys():
+                                row[k] = v
 
-                # If the trial is over (either completed or bailed), flush the row
-                if 'TRIAL_END' in data.keys():
-                    row.append()
-                    table.flush()
-                self.logger.debug('sent data')
+                    # If the trial is over (either completed or bailed), flush the row
+                    if 'TRIAL_END' in data.keys():
+                        row.append()
+                        table.flush()
+                    self.logger.debug('sent data')
 
 
-            # Wait on the stage lock to clear
-            self.stage_block.wait()
-            self.logger.debug('stage lock passed')
+                # Wait on the stage lock to clear
+                self.stage_block.wait()
+                self.logger.debug('stage lock passed')
 
-            # If the running flag gets set, we're closing.
-            if not self.running.is_set():
+                # If the running flag gets set, we're closing.
+                if not self.running.is_set():
+                    break
+
+        except Exception as e:
+            self.logger.exception(f'got exception while running task, task stopping\n {e}')
+
+        finally:
+            self.logger.debug('stopping task')
+            try:
                 self.task.end()
-                self.task = None
-                row.append()
-                table.flush()
-                self.logger.debug('stopping task')
-                break
+            except Exception as e:
+                self.logger.exception(f'got exception while stopping task: {e}')
+            self.task = None
+            row.append()
+            table.flush()
+            gpio.clear_scripts()
+            self.logger.debug('stopped task and cleared scripts')
 
         h5f.flush()
         h5f.close()
